#  Copyright (c) 2022 by Amplo.

import copy
import math
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.metrics import auc, plot_roc_curve
from sklearn.model_selection import StratifiedKFold

from amplo.utils import deprecated
from amplo.utils.logging import get_root_logger

logger = get_root_logger().getChild("Documenter")


@deprecated
class Documenter:
    def __init__(self, pipeline):
        if not os.path.exists(
            pipeline.main_dir + "Documentation/v{}".format(pipeline.version)
        ):
            os.makedirs(
                pipeline.main_dir + "Documentation/v{}".format(pipeline.version)
            )
        self.mName = None
        self.p = pipeline
        self.model = None
        self.params = None
        self.featureSet = None
        self.metrics = None
        self.confusionMatrix = None
        self.x = None
        self.y = None
        self.y_not_normalized = None
        self.cv = None
        self.mode = pipeline.mode
        self.classes_ = None
        if self.mode == "classification" and self.p.y.nunique() > 2:
            self.mode = "multiclass"
            self.classes_ = self.p.y.unique()

    def prepare_data(self):
        """
        Although data of the Pipeline (self.p.x and self.p.y), is already cleaned, it is
        not normalized or sequenced.
        @return:
        @rtype:
        """
        x, y = copy.deepcopy(self.p.x[self.p.colKeep[self.featureSet]]), copy.deepcopy(
            self.p.y
        )

        # Normalize
        if self.p.normalize:
            normalize_features = [
                k for k in x.keys() if k not in self.p.date_cols + self.p.cat_cols
            ]
            x[normalize_features] = self.p.bestScaler.transform(x[normalize_features])
            if self.mode == "regression":
                self.y_not_normalized = y.values.reshape(-1, 1)
                y = pd.Series(
                    self.p.bestOutputScaler.transform(y.values.reshape(-1, 1)).reshape(
                        -1
                    ),
                    name=self.p.target,
                )
        self.x, self.y = x.to_numpy(), y.to_numpy().reshape(-1, 1)
        # todo implement sequencer

    def create(self, model, feature_set):
        """
        Creates Automatic Documentation.

        parameters
        ----------
        model : Estimator
            Model object, requires set_params, fit, predict, predict_proba
        feature_set : str
        """
        self.mName = type(model).__name__
        self.model = model
        self.featureSet = feature_set
        self.prepare_data()
        logger.info(f"{self.mName} feature_set {self.p.version}")

        # Introduction
        markdown = "# Amplo AutoML Documentation - {} v{}\n\n".format(
            self.mName, self.p.version
        )
        markdown += self.model_description()

        # Mode dependent information
        if self.mode == "classification":
            markdown += self.binary_markdown()
        elif self.mode == "multiclass":
            markdown += self.multiclass_markdown()
        elif self.mode == "regression":
            markdown += self.regression_markdown()

        # Cross validation
        markdown += self.validation_strategy_markdown()

        # Parameters
        markdown += self.parameter_markdown()

        # Features
        markdown += self.features_markdown()

        # Data Processing
        markdown += self.data_markdown()

        # Other models
        markdown += self.modelling_markdown()

        return markdown

    def model_description(self):
        markdown = "## Model Information\n\n"
        if "CatBoost" in self.mName:
            markdown += (
                "CatBoost, or **Cat**egorical **Boost**ing,  is an algorithm for gradient boosting on "
                "decision trees, with natural implementation for categorical variables. It is similar to "
                "XGBoost and LightGBM but differs in implementation of the optimization algorithm. We often "
                "see this algorithm performing very well."
            )
        if "XGB" in self.mName:
            markdown += (
                "XGBoost, or E**X**treme **G**radient **Boost**ing, is an algorithm for gradient boosting on "
                "decision trees. It trains many decision trees sequentially, the additional tree always "
                "trying to mitigate the error of the whole model. XGBoost was the first gradient boosting "
                "algorithm to be implemented and is currently widely adopted in the ML world. "
            )
        if "LGBM" in self.mName:
            markdown += (
                "LightGBM, or **L**ight **G**radient **B**oosting **M**achine, is an iteration on the XGBoost"
                " algorithm. Similarly, it uses gradient boosting with decision trees. However, XGBoost tend "
                "to be slow for a larger number of samples (>10.000), but with leaf-wise growth instead of "
                "depth-wise growth, LightGBM increases training speed significatnly. Performance is often "
                "close to XGBoost, sometimes for the better and sometimes for the worse. "
            )
        if "HistGradientBoosting" in self.mName:
            markdown += (
                "SciKits implementation of LightGBM, or **L**ight **G**radient **B**oosting **M**achine, is"
                " an iteration on the XGBoost "
                "algorithm. Similarly, it uses gradient boosting with decision trees. However, XGBoost tend "
                "to be slow for a larger number of samples (>10.000), but with leaf-wise growth instead of "
                "depth-wise growth, LightGBM increases training speed significatnly. Performance is often "
                "close to XGBoost, sometimes for the better and sometimes for the worse. "
            )
        elif "GradientBoosting" in self.mName:
            markdown += (
                "SciKits implementation of XGBoost, or E**X**treme **G**radient Boosting, is an algorithm for"
                " gradient boosting on "
                "decision trees. It trains many decision trees sequentially, the additional tree always "
                "trying to mitigate the error of the whole model. XGBoost was the first gradient boosting "
                "algorithm to be implemented and is currently widely adopted in the ML world. "
            )
        if "RandomForest" in self.mName:
            markdown += (
                "Random Forest, is an ensemble algorithm that combines many (100-1000) decision trees and "
                "predicts the average of all trained trees. Though gradient boosting methods often outperform "
                "Random Forest, some data characteristics favor the Random Forests performance. "
            )
        return markdown + "\n\n"

    def binary_markdown(self):
        # Calculation
        accuracy = []
        precision = []
        sensitivity = []
        specificity = []
        f1_score = []
        area_under_curves = []
        true_positive_rates = []
        cm = np.zeros((self.p.cv_splits, 2, 2))
        mean_fpr = np.linspace(0, 1, 100)

        # Plot Initiator
        fig, ax = plt.subplots(
            math.ceil(self.p.cv_splits / 2),
            2,
            sharex="all",
            sharey="all",
            figsize=[12, 8],
        )
        fig.suptitle(
            "{}-Fold Cross Validated Predictions - {} ({})".format(
                self.p.cv_splits, type(self.model).__name__, self.featureSet
            )
        )
        fig2, ax2 = plt.subplots(figsize=[12, 8])

        # Modelling
        self.cv = StratifiedKFold(n_splits=self.p.cv_splits, shuffle=self.p.shuffle)
        for i, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            n = len(v)
            xt, xv, yt, yv = (
                self.x[t],
                self.x[v],
                self.y[t].reshape((-1)),
                self.y[v].reshape((-1)),
            )
            model = copy.deepcopy(self.model)
            model.fit(xt, yt)
            predictions = model.predict(xv).reshape((-1))

            # Metrics
            tp = np.logical_and(np.sign(predictions) == 1, yv == 1).sum()
            tn = np.logical_and(np.sign(predictions) == 0, yv == 0).sum()
            fp = np.logical_and(np.sign(predictions) == 1, yv == 0).sum()
            fn = np.logical_and(np.sign(predictions) == 0, yv == 1).sum()
            accuracy.append((tp + tn) / n * 100)
            if tp + fp > 0:
                precision.append(tp / (tp + fp) * 100)
            if tp + fn > 0:
                sensitivity.append(tp / (tp + fn) * 100)
            if tn + fp > 0:
                specificity.append(tn / (tn + fp) * 100)
            if tp + fp > 0 and tp + fn > 0:
                f1_score.append(
                    2
                    * precision[-1]
                    * sensitivity[-1]
                    / (precision[-1] + sensitivity[-1])
                    if precision[-1] + sensitivity[-1] > 0
                    else 0
                )
            cm[i] = np.array([[tp, fp], [fn, tn]])

            # ROC calculations
            viz = plot_roc_curve(
                model, xv, yv, name="ROC fold {}".format(i + 1), alpha=0.3, ax=ax2
            )
            interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
            interp_tpr[0] = 0.0
            true_positive_rates.append(interp_tpr)
            area_under_curves.append(viz.roc_auc)

            # Plot
            ax[i // 2][i % 2].plot(yv, c="#2369ec", alpha=0.6)
            ax[i // 2][i % 2].plot(predictions, c="#ffa62b")
            ax[i // 2][i % 2].set_title("Fold-{}".format(i))

        # Statistics on results
        totals = np.sum(cm, axis=(1, 2), keepdims=True)
        means = np.mean(cm / totals * 100, axis=0)
        stds = np.std(cm / totals * 100, axis=0)

        # Print Results
        logger.info(
            f"Accuracy:        {np.mean(accuracy):.2f} \u00B1 {np.std(accuracy):.2f} %"
        )
        logger.info(
            f"Precision:       {np.mean(precision):.2f} \u00B1 {np.std(precision):.2f} %"
        )
        logger.info(
            f"Recall:          {np.mean(sensitivity):.2f} \u00B1 {np.std(sensitivity):.2f} %"
        )
        logger.info(
            f"Specificity:     {np.mean(specificity):.2f} \u00B1 {np.std(specificity):.2f} %"
        )
        logger.info(
            f"F1-score:        {np.mean(f1_score):.2f} \u00B1 {np.std(f1_score):.2f} %"
        )
        logger.info("Confusion Matrix:")
        logger.info("Prediction / true |    Faulty    |    Healthy      ")
        s1 = f"{means[0, 0]:.1f} \u00B1 {stds[0, 0]:.1f} %".ljust(12)
        s2 = f"{means[0, 1]:.1f} \u00B1 {stds[0, 1]:.1f} %".ljust(12)
        logger.info(f"      Faulty      | {s1} |  {s2}")
        s1 = f"{means[1, 0]:.1f} \u00B1 {stds[1, 0]:.1f} %".ljust(12)
        s2 = f"{means[1, 1]:.1f} \u00B1 {stds[1, 1]:.1f} %".ljust(12)
        logger.info(f"      Healthy     | {s1} |  {s2}")

        # Check whether plot is possible
        if type(self.model).__name__ == "Lasso" or "Ridge" in type(self.model).__name__:
            return

        # Adjust plots
        ax2.plot(
            [0, 1],
            [0, 1],
            linestyle="--",
            lw=2,
            color="#ffa62b",
            label="Chance",
            alpha=0.8,
        )
        mean_tpr = np.mean(true_positive_rates, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = auc(mean_fpr, mean_tpr)
        std_auc = np.std(area_under_curves)
        ax2.plot(
            mean_fpr,
            mean_tpr,
            color="#2369ec",
            label=r"Mean ROC (AUC = {:.2f} $\pm$ {:.2f})".format(mean_auc, std_auc),
            lw=2,
            alpha=0.8,
        )
        std_tpr = np.std(true_positive_rates, axis=0)
        true_pos_rates_upper = np.minimum(mean_tpr + std_tpr, 1)
        true_pos_rates_lower = np.maximum(mean_tpr - std_tpr, 0)
        ax2.fill_between(
            mean_fpr,
            true_pos_rates_lower,
            true_pos_rates_upper,
            color="#729ce9",
            alpha=0.2,
            label=r"$\pm$ 1 std. dev.",
        )
        ax2.set(
            xlim=[-0.05, 1.05],
            ylim=[-0.05, 1.05],
            title="ROC Curve - {}".format(self.mName),
        )
        ax2.legend(loc="lower right")
        roc_path = self.p.main_dir + "Documentation/v{}/ROC_{}.png".format(
            self.p.version, self.mName
        )
        fig2.savefig(roc_path, format="png", dpi=200)
        cross_val_path = self.p.main_dir + "Documentation/v{}/Cross_Val_{}.png".format(
            self.p.version, self.mName
        )
        fig.savefig(cross_val_path, format="png", dpi=200)

        # Markdown
        metrics_table = """| Metric | Score |
| --- | ---: |
| Accuracy    | {:.2f} \u00B1 {:.2f} % |
| Precision   | {:.2f} \u00B1 {:.2f} % |
| Sensitivity | {:.2f} \u00B1 {:.2f} % |
| Specificity | {:.2f} \u00B1 {:.2f} % |
| F1 Score    | {:.2f} \u00B1 {:.2f} % |""".format(
            np.mean(accuracy),
            np.std(accuracy),
            np.mean(precision),
            np.std(accuracy),
            np.mean(sensitivity),
            np.std(sensitivity),
            np.mean(specificity),
            np.std(specificity),
            np.mean(f1_score),
            np.std(f1_score),
        )
        confusion_matrix = """
<table>
    <thead>
        <tr>
            <td> </td>
            <td> </td>
            <td colspan=2 style="text-align:center">
                True Label
            </td>
        </tr>
        <tr>
            <td> </td>
            <td> </td>
            <td>Faulty</td>
            <td>No Issue</td>
        </tr>
    </thead>
    <tbody>
        <tr>
            <td rowspan=2 style="vertical-align:middle">
                Prediction
            </td>
            <td>Faulty</td>
            <td>{:.1f} \u00B1 {:.1f} %</td>
            <td>{:.1f} \u00B1 {:.1f} %</td>
        </tr>
        <tr>
            <td>No Issue</td>
            <td>{:.1f} \u00B1 {:.1f} %</td>
            <td>{:.1f} \u00B1 {:.1f} %</td>
        </tr>
    </tbody>
</table>
        """.format(
            means[0, 0],
            stds[0, 0],
            means[0, 1],
            stds[0, 1],
            means[1, 0],
            stds[1, 0],
            means[1, 1],
            stds[1, 1],
        )
        return """## Model Performance

Model performance is analysed by various metrics. Below you find various metrics and a confusion matrix.
This model has been selected based on the {} score.

### Metrics

{}

### Confusion Matrix

{}

### Area Under Curve & Cross Validation Plot

<table>
    <tr>
        <td><img src="{}" width=600 height=400></td>
        <td><img src="{}" width=600 height=400></td>
    </tr>
<table>

""".format(
            self.p.objective,
            metrics_table,
            confusion_matrix,
            "ROC_{}.png".format(self.mName),
            "Cross_Val_{}.png".format(self.mName),
        )

    def multiclass_markdown(self):
        # Initiating
        fig, ax = plt.subplots(
            math.ceil(self.p.cv_splits / 2), 2, sharex="all", sharey="all"
        )
        fig.suptitle(
            "{}-Fold Cross Validated Predictions - {} ({})".format(
                self.p.cv_splits, self.mName, self.featureSet
            )
        )
        n_classes = len(self.classes_)
        f1_score = np.zeros((self.p.cv_splits, n_classes))
        log_loss = np.zeros(self.p.cv_splits)
        avg_acc = np.zeros(self.p.cv_splits)
        cm = np.zeros((self.p.cv_splits, n_classes, n_classes))

        # Modelling
        self.cv = StratifiedKFold(n_splits=self.p.cv_splits, shuffle=self.p.shuffle)
        for i, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            xt, xv, yt, yv = (
                self.x[t],
                self.x[v],
                self.y[t].reshape((-1)),
                self.y[v].reshape((-1)),
            )
            model = copy.copy(self.model)
            model.fit(xt, yt)
            predictions = model.predict(xv).reshape((-1))
            cm[i] = metrics.confusion_matrix(predictions, yv)

            # Metrics
            f1_score[i] = metrics.f1_score(yv, predictions, average=None)
            avg_acc[i] = metrics.accuracy_score(yv, predictions)
            if hasattr(model, "predict_proba"):
                probabilities = model.predict_proba(xv)
                log_loss[i] = metrics.log_loss(yv, probabilities)

            # Plot
            ax[i // 2][i % 2].plot(yv, c="#2369ec", alpha=0.6)
            ax[i // 2][i % 2].plot(predictions, c="#ffa62b")
            ax[i // 2][i % 2].set_title("Fold-{}".format(i))

        # Results
        logger.info("F1 scores:")
        logger.info("".join([" Class {} |".format(i) for i in range(n_classes)]))
        logger.info(
            "".join(
                [
                    " {:.2f} % ".ljust(11).format(f1) + "|"
                    for f1 in np.mean(f1_score, axis=0)
                ]
            )
        )
        logger.info(
            "Average Accuracy: {:.2f} \u00B1 {:.2f} %".format(
                np.mean(avg_acc), np.std(avg_acc)
            )
        )
        if hasattr(model, "predict_proba"):
            logger.info(
                "Log Loss:         {:.2f} \u00B1 {:.2f}".format(
                    np.mean(log_loss), np.std(log_loss)
                )
            )

        # Result statistics
        totals = np.sum(cm, axis=(1, 2), keepdims=True)
        means = np.mean(cm / totals * 100, axis=0)
        stds = np.std(cm / totals * 100, axis=0)

        # Markdown
        metrics_table = """| Metrcis | Score |
| --- | ---: |
| Avg. Accuracy | {:.2f} \u00B1 {:.2f} |
| F1 Score      | {:.2f} \u00B1 {:.2f} |
""".format(
            np.mean(avg_acc), np.std(avg_acc), np.mean(f1_score), np.std(f1_score)
        )
        if hasattr(model, "predict_proba"):
            metrics_table += "| Log Loss      | {:.2f} \u00B1 {:.2f} |".format(
                np.mean(log_loss), np.std(log_loss)
            )
        classes_header = "<tr><td> </td><td> </td>{}</tr>" "".format(
            "".join(
                ["<td> {} </td>".format(self.classes_[i]) for i in range(n_classes)]
            )
        )
        prediction = (
            '<td rowspan={} style="vertical-align:middle">Prediction</td>'.format(
                n_classes
            )
        )
        rows = "\n".join(
            [
                "<tr>\n{}\n{}\n{}\n</tr>".format(
                    prediction if i == 0 else "",
                    "<td>\n{}\n</td>".format(self.classes_[i]),
                    "\n".join(
                        [
                            "<td>\n{:.2f} \u00B1 {:.2f}\n</td>".format(
                                means[j, i], stds[j, i]
                            )
                            for j in range(n_classes)
                        ]
                    ),
                )
                for i in range(n_classes)
            ]
        )
        confusion_matrix = """
<table>
    <thead>
        <tr>
            <td> </td>
            <td> </td>
            <td colspan={} style="text-align:center">True Label</td>
        </tr>
        {}
    </thead>
    <tbody>
        {}
    </tbody>
</table>
                
""".format(
            n_classes, classes_header, rows
        )
        return """## Model Performance

Model performance is analysed by various metrics. Below you find various metrics and a confusion matrix.
This model has been selected based on the {} score.

### Metrics

{}

### Confusion Matrix

{}

""".format(
            self.p.objective, metrics_table, confusion_matrix
        )

    def regression_markdown(self):
        return "## Model Performance n/a"

    #         return """## Model Performance
    #
    #
    # Model performance is analysed by various metrics. Below you find various metrics and a cross validated prediction plot.
    #
    # ### Metrics
    #
    # {}
    #
    # ### Prediction Plot
    #
    # <img src="Cross_Val_{}.png" width=1200, height=400>
    #
    # """.format(metrics_table, self.mName)

    def validation_strategy_markdown(self):
        return """## Validation Strategy

All experiments are cross validated. This means that every time a model"s performance is evaluated, it"s trained on one part of the data, and tested on another. Therefore, the model is always tested against data it has not yet been trained for. This gives the best approximation for real world (out of sample) performance.
The current validation strategy used {}, with {} splits and {} shuffling the data.
""".format(
            type(self.cv).__name__,
            self.p.cv_splits,
            "with" if self.p.shuffle else "without",
        )

    @staticmethod
    def split_param_table(params):
        if len(params) > 10:
            table = ""
            for i in range(-(-len(params) // 10)):  # Ceil
                table += (
                    "| "
                    + " | ".join(
                        [k for k in list(params.keys())[i * 10 : (i + 1) * 10]]
                    )
                    + "|"
                )
                table += "\n|" + " --- |" * min(len(params) - i * 10, 10)
                table += (
                    "\n| "
                    + " | ".join(
                        [
                            str(v) if not isinstance(v, float) else str(round(v, 4))
                            for v in list(params.values())[i * 10 : (i + 1) * 10]
                        ]
                    )
                    + " |\n\n"
                )
        else:
            table = "| " + " | ".join([k for k in params.keys()]) + " |"
            table += "\n|" + " --- |" * 10
            table += (
                "\n| "
                + " | ".join(
                    [
                        str(v) if not isinstance(v, float) else str(round(v, 4))
                        for v in params.values()
                    ]
                )
                + " |\n\n"
            )
        return table

    def parameter_markdown(self):
        # Parameter section for stacking
        if "Stacking" in self.mName:
            estimators = self.model.get_params()["estimators"]
            table = ""
            for name, estimator in estimators:
                params = estimator.get_params()
                table += self.split_param_table(params)
        # Other table
        else:
            params = self.model.get_params()
            table = self.split_param_table(params)
        return """## Parameters

The optimized model has the following parameters:

{}

""".format(
            table
        )

    def features_markdown(self):
        # Calculate feature importance
        feature_importance_url = (
            self.p.main_dir
            + "Documentation/v{}/Feature_Importance_{}.png".format(
                self.p.version, self.mName
            )
        )
        if not os.path.exists(feature_importance_url):
            rf = None
            if self.mode == "regression":
                rf = RandomForestRegressor().fit(self.p.x, self.p.y)
            else:
                rf = RandomForestClassifier().fit(self.p.x, self.p.y)
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines["right"].set_visible(False)
            ax.spines["bottom"].set_visible(False)
            ax.spines["top"].set_visible(False)
            ind = np.argsort(rf.feature_importances_)
            plt.barh(
                list(self.p.x.keys()[ind])[-15:],
                width=rf.feature_importances_[ind][-15:],
                color="#2369ec",
            )
            fig.savefig(feature_importance_url, format="png", dpi=200)

        # Number of Features
        n_co_linear = len(self.p.feature_processor.coLinearFeatures)
        n_cross = len(self.p.feature_processor.crossFeatures)
        n_additive = len(self.p.feature_processor.addFeatures)
        n_trigonometry = len(self.p.feature_processor.trigoFeatures)
        n_k_means = len(self.p.feature_processor.kMeansFeatures)
        n_lagged = len(self.p.feature_processor.laggedFeatures)
        n_diff = len(self.p.feature_processor.diffFeatures)
        return """## Features

### Feature Extraction

Firstly, features that are co-linear (a * x = y), up to {} %, were removed. This resulted in {} removed features:
{} 

Subsequently, the features were manipulated and analysed to extract additional information. 
Most combinations are tried (early stopping avoids unpromising features directly). 
The usefulness of a newly extracted features are analysed by a single decision tree. 

| Sort Feature | Quantity | Features |
| --- | ---: | --- |
| Multiplied / Divided Features | {} | {} |
| Added / Subtracted Features   | {} | {} |
| Trigonometric Features        | {} | {} |
| K-Means Features              | {} | {} |
| Lagged Features               | {} | {} |
| Differentiated Features       | {} | {} |

### Feature Selection
Using a Random Forest model, the non-linear Feature Importance is analysed. The Feature Importance is measured
in Mean Decrease in Gini Impurity. 
The Feature Importance is used to create two feature sets, one that contains 95% of all Feature Importance (RFT) and 
one that contains all features that contribute more than 1% to the total Feature Importance (RFI). 

Top 20 features:

<img src="{}" width="400" height="600">
""".format(
            self.p.information_threshold * 100,
            n_co_linear,
            ", ".join([i for i in self.p.feature_processor.coLinearFeatures]),
            n_cross,
            ", ".join([i for i in self.p.feature_processor.crossFeatures]),
            n_additive,
            ", ".join([i for i in self.p.feature_processor.addFeatures]),
            n_trigonometry,
            ", ".join([i for i in self.p.feature_processor.trigoFeatures]),
            n_k_means,
            ", ".join([i for i in self.p.feature_processor.kMeansFeatures]),
            n_lagged,
            ", ".join([i for i in self.p.feature_processor.laggedFeatures]),
            n_diff,
            ", ".join([i for i in self.p.feature_processor.diffFeatures]),
            "Feature_Importance_{}.png".format(self.mName),
        )

    def data_markdown(self):
        return """## Data Processing
        
Data cleaning steps: 
1. Removed {} duplicate columns and {} duplicate rows.
2. Handled outliers with {}
3. Imputed {} missing values with {}
4. Removed {} columns with only constant values

""".format(
            self.p.data_processor.removedDuplicateColumns,
            self.p.data_processor.removedDuplicateRows,
            self.p.outlier_removal,
            self.p.data_processor.imputedMissingValues,
            self.p.missing_values,
            self.p.data_processor.removedConstantColumns,
        )

    def modelling_markdown(self):
        top_ten_table = "| Model | {} | Parameters |\n| --- | ---: | --- |\n".format(
            self.p.objective.ljust(16)
        )
        for i in range(min(10, len(self.p.results))):
            model = self.p.results.iloc[i]["model"]
            score = "{:.4f} \u00B1 {:.4f}".format(
                self.p.results.iloc[i]["mean_objective"],
                self.p.results.iloc[i]["std_objective"],
            ).ljust(16)
            params = self.p.results.iloc[i]["params"]
            top_ten_table += "| {} | {} | {} |\n".format(model, score, params)
        return """## Model Score Board

Not only {} has been optimized by the AutoML pipeline. In total, {} models were trained. 
The following table shows the performance of the top 10 performing models:

{}

""".format(
            self.mName, len(self.p.results), top_ten_table
        )



#  Copyright (c) 2022 by Amplo.

"""
Base exceptions.
"""

__all__ = [
    "ExperimentalWarning",
    "NotFittedError",
]


# --------------------------------------------------------------------------
# Warnings


class ExperimentalWarning(UserWarning):
    """
    Warning for experimental features.
    """


# --------------------------------------------------------------------------
# Errors


class EmptyFileError(Exception):
    """
    File contains no data.
    """


class NotFittedError(Exception):
    """
    Object is not fitted.
    """



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import io
import json
import os
import warnings
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pandas as pd
import pytz
from azure.storage.blob import BlobServiceClient

from amplo.base.exceptions import EmptyFileError
from amplo.utils.util import check_dtypes, deprecated

if TYPE_CHECKING:
    from azure.storage.blob import BlobClient, ContainerClient

__all__ = ["AzureSynchronizer"]


_AZURE_CLIENT_NAME = "amploplatform"
_AZURE_CONNECTION_STR_OS = "AZURE_STORAGE_STRING"


class AzureBlobDataAPI:
    """
    Helper class for handling data from an Azure blob storage.

    Parameters
    ----------
    client : str
        Container client name.
    connection_str : str
        Connection string for given container client.
    """

    def __init__(self, client: str, connection_str: str):
        check_dtypes(("client", client, str), ("connection_str", connection_str, str))
        self.client = client
        self.connection_str = connection_str
        bsc = BlobServiceClient.from_connection_string(connection_str)
        self._container: ContainerClient = bsc.get_container_client(client)

    def __repr__(self):
        """
        Readable string representation of the class.
        """

        return f"{self.__class__.__name__}({self.client})"

    @classmethod
    def from_os_env(
        cls, client: str | None = None, connection_str_os: str | None = None
    ) -> AzureBlobDataAPI:
        """
        Instantiate the class using os environment strings.

        Parameters
        ----------
        client : str, optional, default: _AZURE_CLIENT_NAME
            Container client name.
        connection_str_os : str, optional, default: _AZURE_CONNECTION_STR_OS
            Key in the os environment for the Azure connection string.

        Returns
        -------
        AzureBlobDataHandler

        Raises
        ------
        KeyError
            When a os variable is not set.
        """

        connection_str_os = connection_str_os or _AZURE_CONNECTION_STR_OS

        client = client or _AZURE_CLIENT_NAME
        connection_str = os.environ[connection_str_os]
        return cls(client, connection_str)

    # --------------------------------------------------------------------------
    # Blob inspection

    def ls(self, path: str | Path | None = None) -> list[str]:

        # Provide slash from right
        if path is not None:
            path = f"{Path(path).as_posix()}/"

        # List all files and folders
        return [f.name for f in self._container.walk_blobs(path)]  # type: ignore

    def ls_files(self, path: str | Path | None = None) -> list[str]:

        return [f for f in self.ls(path) if not f.endswith("/")]

    def ls_folders(self, path: str | Path | None = None) -> list[str]:

        return [f for f in self.ls(path) if f.endswith("/")]

    # --------------------------------------------------------------------------
    # File handling

    def get_blob_client(self, path: str | Path) -> BlobClient:
        # Check input
        if isinstance(path, Path):
            path = str(path.as_posix())

        # Get blob client
        return self._container.get_blob_client(path)

    def get_metadata(self, path: str | Path) -> dict[str, str | float | int]:

        props = self.get_blob_client(path).get_blob_properties()
        return {  # type: ignore
            "file_name": Path(props.name).name,  # type: ignore
            "full_path": props.name,
            "container": props.container,
            # "creation_time": props.creation_time.timestamp(),
            "last_modified": props.last_modified.timestamp(),
        }

    def get_size(self, path: str | Path) -> int:

        return self.get_blob_client(path).get_blob_properties().size  # type: ignore

    def download_file(
        self, path: str | Path, local_path: str | Path, match_timestamps: bool = True
    ) -> None:

        blob = self.get_blob_client(path)

        # Ensure that local_path's parent exists
        Path(local_path).parent.mkdir(parents=True, exist_ok=True)
        if isinstance(local_path, Path):
            local_path = str(local_path.as_posix())

        # Write blob data to local_path
        with open(local_path, "wb") as f:
            f.write(blob.download_blob().readall())

        # Manipulate file properties
        if match_timestamps:
            properties = blob.get_blob_properties()
            created: float = properties.creation_time.timestamp()
            last_modified: float = properties.last_modified.timestamp()
            os.utime(local_path, (created, last_modified))

    def read_json(self, path: str | Path) -> list | dict:

        blob = self.get_blob_client(path)
        return json.loads(blob.download_blob().readall())

    def read_pandas(self, path: str | Path, **kwargs) -> pd.Series | pd.DataFrame:

        from amplo.utils.io import FILE_READERS

        # Check whether a proper file reader exists for the file
        file_extension = Path(path).suffix
        pandas_reader = FILE_READERS.get(file_extension)
        if pandas_reader is None:
            raise NotImplementedError(f"File format {file_extension} not supported.")

        # Read buffered data into pandas
        blob = self.get_blob_client(path)
        file_buffer = io.BytesIO(blob.download_blob().readall())
        return pandas_reader(file_buffer, **kwargs)


@deprecated("This class won't be updated anymore.")
class AzureSynchronizer:
    def __init__(
        self,
        connection_string_name: str = "AZURE_STORAGE_STRING",
        container_client_name: str = "amploplatform",
        verbose=0,
    ):
        """
        Connector to Azure storage blob for downloading data that is stored
        in Amplo`s data storage fashion.

        Parameters
        ----------
        connection_string_name : str
        container_client_name : str
        verbose : int
        """
        client = BlobServiceClient.from_connection_string(
            os.getenv(connection_string_name, "unkown_azure_connection_string")
        )
        self.container = client.get_container_client(container_client_name)
        self.verbose = int(verbose)
        self._metadata_filename = ".metadata"
        self._str_time_format = "%Y-%m-%d %H:%M:%S:%z"

    def get_dir_paths(self, path=None):
        """
        Get all directories that are direct children of given directory (``path``).

        Parameters
        ----------
        path : str or Path, optional
            Path to search for directories.
            If not provided, searches in root `/`.

        Returns
        -------
        list of str
        """
        if path is not None:
            # Provide a slash from right
            path = f"{Path(path).as_posix()}/"
        dirs = [
            b.name for b in self.container.walk_blobs(path) if str(b.name).endswith("/")
        ]
        return dirs

    def get_filenames(self, path, with_prefix=False, sub_folders=False):
        """
        Get all files that are direct children of given directory (``path``).

        Parameters
        ----------
        path : str or Path
            Path to search for files
        with_prefix : bool
            Whether to fix the prefix of the files
        sub_folders : bool
            Whether to search also for files inside sub-folders

        Returns
        -------
        list of str
        """
        # Provide a slash from right
        path = f"{Path(path).as_posix()}/"

        # List files
        if sub_folders:
            files = [
                f.name
                for f in self.container.walk_blobs(path, delimiter="")
                if "." in f.name
            ]
        else:
            files = [
                f.name
                for f in self.container.walk_blobs(path, delimiter="")
                if f.name.count("/") == path.count("/") and "." in f.name
            ]

        # Fix prefix
        if not with_prefix:
            files = [f[len(path) :] for f in files]

        # Remove empties
        if "" in files:
            files.remove("")

        return files

    def sync_files(self, blob_dir, local_dir):
        """
        Download all files inside blob directory and store it to the local directory.

        Additionally, creates a file `.metadata` that stores additional info about
        synchronization such as blob directory path and last modification date.

        Parameters
        ----------
        blob_dir : str or Path
            Search directory (download)
        local_dir : str or Path
            Local directory (store)

        Returns
        -------
        found_new_data : bool
            Whether new data has been downloaded

        Notes
        -----
        The data in the `.metadata` file is currently only used for checking the last
        modification date, thus telling whether files have to be downloaded / updated.
        """
        # Set up paths
        blob_dir = Path(blob_dir)
        local_dir = Path(local_dir)
        metadata_dir = local_dir

        # Warn when names don't match
        if blob_dir.name != local_dir.name:
            warnings.warn(
                f"Name mismatch detected. {blob_dir.name} != {local_dir.name}"
            )

        # Skip "Random"
        if blob_dir.name == "Random":
            warnings.warn(f"Skipped synchronization from {blob_dir}")
            return False

        # Set up metadata
        blob_dir_properties = self.container.get_blob_client(
            str(blob_dir)
        ).get_blob_properties()
        metadata: dict[str, str | list[str] | datetime] = dict(
            blob_dir=str(blob_dir),
            container=blob_dir_properties.container,
            last_modified=blob_dir_properties.last_modified,
            new_files=[],
        )
        metadata["last_modified"] = cast(datetime, metadata["last_modified"])
        metadata["new_files"] = cast(list[str], metadata["new_files"])

        # Load local metadata from previous synchronization
        local_metadata = self.load_local_metadata(metadata_dir, not_exist_ok=True)
        last_updated = local_metadata.get(
            "last_modified", datetime(1900, 1, 1, tzinfo=pytz.UTC)
        )

        # Read & write all files
        for file in self.get_filenames(blob_dir):

            # Create directory only if files are found
            local_dir.mkdir(parents=True, exist_ok=True)

            # Get file blob
            blob = self.container.get_blob_client(str(blob_dir / file))
            blob_properties = blob.get_blob_properties()

            # Download and save if file is new or modified
            file_created: datetime = blob_properties.creation_time
            file_last_modified: datetime = blob_properties.last_modified
            if file_last_modified > last_updated:
                # Write file
                with open(str(local_dir / file), "wb") as f:
                    f.write(blob.download_blob().readall())
                # Match timestamps of local file with blob
                os.utime(
                    str(local_dir / file),
                    (file_created.timestamp(), file_last_modified.timestamp()),
                )
                # Increment
                metadata["new_files"].append(str(file))

        # Check whether found new data
        found_new_data = metadata["last_modified"] > last_updated
        if found_new_data:
            # Store metadata
            self._dump_local_metadata(metadata, metadata_dir)

        return found_new_data

    # --- Utilities ---

    def load_local_metadata(self, local_dir, *, not_exist_ok=False):
        metadata_path = Path(local_dir) / self._metadata_filename
        if metadata_path.exists():
            # Get and check local metadata
            metadata = json.load(open(str(metadata_path), "r"))
            assert isinstance(metadata, dict), f"Damaged metadata in {metadata_path}"
            # Convert string to datetime object
            metadata["last_modified"] = datetime.strptime(
                metadata["last_modified"], self._str_time_format
            )
            return metadata
        elif not_exist_ok:
            return dict()
        else:
            raise FileNotFoundError(f"File {metadata_path} does not exist")

    def _dump_local_metadata(self, metadata, local_dir):
        metadata_path = Path(local_dir) / self._metadata_filename
        # Check metadata keys
        assert {"blob_dir", "container", "new_files", "last_modified"}.issubset(
            metadata.keys()
        ), "Invalid metadata"
        # Make datetime object JSON serializable
        metadata["last_modified"] = datetime.strftime(
            metadata["last_modified"], self._str_time_format
        )
        # Dump
        json.dump(metadata, open(str(metadata_path), "w"))



#  Copyright (c) 2022 by Amplo.

import time

import numpy as np
import pandas as pd
from imblearn.combine import SMOTEENN, SMOTETomek
from imblearn.over_sampling import SMOTE, BorderlineSMOTE, RandomOverSampler
from imblearn.under_sampling import (
    EditedNearestNeighbours,
    OneSidedSelection,
    RandomUnderSampler,
    TomekLinks,
)
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import StratifiedKFold

from amplo.utils.logging import get_root_logger

logger = get_root_logger().getChild("DataSampler")


class DataSampler:
    def __init__(
        self,
        method: str = "both",
        margin: float = 0.1,
        cv_splits: int = 3,
        shuffle: bool = True,
        fast_run: bool = False,
        objective: str = "neg_log_loss",
        verbosity: int = 1,
    ):
        """
        Trial and error various strategies:
        - Under-sample all to minority
        - Over-sample all to majority
        - Combination of the two, with say 1k, 5k, 25k, 50k, 75k and 100k samples
        This sensitivity is problem dependent, but has major consequence for hyper
        parameter optimization. Need to get it right.
        Need to make sure that metrics are evaluated on original distribution.
        https://imbalanced-learn.org/stable/references/index.html#api

        Parameters
        ----------
        method : str
            Whether to under-sample, over-sample, or a combination from the both
            ("under", "over", or "both")
        margin : float
            If the imbalance is smaller than this margin, skip the balancing, should be
            within [0, 1]
        fast_run : bool
            If true, skips exhaustive samplers
        objective : str
            Scikit-Learn metric
        """
        # Tests
        if method not in ("under", "over", "both"):
            raise ValueError("Method must be one of 'under', 'over' or 'both'.")
        if not (1e-6 < margin < 1):
            raise ValueError("Margin must be within [1e-6, 1].")

        # Objective
        self.objective = objective or "neg_log_loss"
        self.scorer = metrics.get_scorer(self.objective)

        # Parse
        self.method = method
        self.margin = margin
        self.fast_run = fast_run

        # Settings
        self.cv = StratifiedKFold(n_splits=cv_splits, shuffle=shuffle)
        self.min_samples = 2e2
        self.max_samples = 1e5
        self.verbosity = verbosity

        # Results
        self.is_fitted = False
        self.opt_sampler = None
        self.results = pd.DataFrame(
            columns=[
                "sampler",
                "mean_objective",
                "std_objective",
                "worst_case",
                "mean_time",
                "std_time",
            ]
        )

    def get_samplers(self, y: pd.Series) -> list:
        """
        Return selection of samplers based on the label vector
        Parameters
        ----------
        labels [pd.Series]: Vector with labels
        """
        # Assess labels
        n_classes = y.nunique(dropna=True)
        n_samples = len(y)
        imbalance = y.value_counts().values / n_samples
        n_min_samples = y.value_counts().min() - 2
        logger.info(
            "Data Balancer: Found class balance: {}.".format(
                ", ".join(["{:.1f}%".format(i * 100) for i in imbalance])
            )
        )

        # If imbalance is negligible, return nothing
        if np.min(imbalance) > (1 - self.margin) / n_classes:
            return []

        # Init
        samplers = []

        # Under Samplers
        if self.method != "over" and n_samples > self.min_samples:
            samplers.append(
                OneSidedSelection(
                    sampling_strategy="not minority", n_seeds_S=int(n_samples / 10)
                )
            )
            samplers.append(RandomUnderSampler(sampling_strategy="not minority"))
            if not self.fast_run:
                samplers.append(TomekLinks(sampling_strategy="not minority"))

        # Over Samplers
        if self.method != "under" and n_samples < self.max_samples:
            samplers.append(RandomOverSampler(sampling_strategy="not majority"))
            if not self.fast_run:
                samplers.append(
                    SMOTE(
                        sampling_strategy="not majority",
                        k_neighbors=min(n_min_samples - 1, 6),
                    )
                )
                samplers.append(
                    BorderlineSMOTE(
                        sampling_strategy="not majority",
                        k_neighbors=min(n_min_samples - 1, 5),
                        m_neighbors=min(n_min_samples, 10),
                    )
                )

        if (
            self.method == "both"
            and self.min_samples < n_samples < self.max_samples
            and not self.fast_run
        ):
            samplers.append(
                SMOTETomek(
                    sampling_strategy="all",
                    smote=SMOTE(
                        sampling_strategy="all", k_neighbors=min(n_min_samples, 6)
                    ),
                )
            )
            samplers.append(
                SMOTEENN(
                    sampling_strategy="all",
                    enn=EditedNearestNeighbours(
                        sampling_strategy="all", n_neighbors=min(n_min_samples, 3)
                    ),
                    smote=SMOTE(
                        sampling_strategy="all", k_neighbors=min(n_min_samples, 6)
                    ),
                )
            )

        return samplers

    def _score(self, sampler, x: pd.DataFrame, y: pd.Series) -> [list, list]:
        """
        Scores a sampler for data

        Parameters
        ----------
        sampler : Imbalance-learn sampler
        x [pd.DataFrame]: Input data
        y [pd.Series]: Output data

        Returns
        -------
        Scores [list]: List of objective for every fold
        Times [list]: List of runtimes for every fold
        """
        # Initiate
        scores, times = [], []

        # Cross-Validate
        for i, (ti, vi) in enumerate(self.cv.split(x, y)):
            # Split data
            xt, xv, yt, yv = x.iloc[ti], x.iloc[vi], y.iloc[ti], y.iloc[vi]

            # Note start time
            t_start = time.time()

            # Re - sample
            if sampler is not None:
                x_res, y_res = sampler.fit_resample(xt, yt)
            else:
                x_res, y_res = xt, yt

            # Train a model
            model = RandomForestClassifier(n_estimators=50, max_depth=5)
            model.fit(x_res, y_res)

            # Score
            scores.append(self.scorer(model, xv, yv))
            times.append(time.time() - t_start)

        return scores, times

    def fit(self, x: pd.DataFrame, y: pd.Series):
        """

        Parameters
        ----------
        x [pd.DataFrame]: Input data
        y [pd.Series]: Output data
        """
        # Get samplers
        samplers = self.get_samplers(y)

        # If samplers are empty, we don"t need sampling
        if len(samplers) == 0:
            self.opt_sampler = None

        else:
            if self.verbosity > 0:
                logger.info("Balancing data with {:.0f} samplers".format(len(samplers)))

            # Get baseline
            scores, times = self._score(None, x, y)
            self.results = self.results.append(
                {
                    "sampler": "Baseline",
                    "mean_objective": np.mean(scores),
                    "std_objective": np.std(scores),
                    "worst_case": np.mean(scores) - np.std(scores),
                    "mean_time": np.mean(times),
                    "std_time": np.std(scores),
                },
                ignore_index=True,
            )
            if self.verbosity > 0:
                logger.info(
                    "Fitted {} {}: {:.4f} \u00B1 {:.4f}".format(
                        "Baseline".ljust(25),
                        self.objective,
                        np.mean(scores),
                        np.std(scores),
                    )
                )

            # Iterate through samplers
            for sampler in samplers:
                name = type(sampler).__name__

                # Score
                scores, times = self._score(sampler, x, y)

                # Store results
                self.results = self.results.append(
                    {
                        "sampler": name,
                        "mean_objective": np.mean(scores),
                        "std_objective": np.std(scores),
                        "worst_case": np.mean(scores) - np.std(scores),
                        "mean_time": np.mean(times),
                        "std_time": np.std(times),
                    },
                    ignore_index=True,
                )

                # Print results
                if self.verbosity > 0:
                    logger.info(
                        "Fitted {} {}: {:.4f} \u00B1 {:.4f}".format(
                            name.ljust(25),
                            self.objective,
                            np.mean(scores),
                            np.std(scores),
                        )
                    )

            # Set optimal sampler
            opt_sampler = self.results.loc[
                self.results["worst_case"] == self.results["worst_case"].max(),
                "sampler",
            ].iloc[0]
            if opt_sampler != "Baseline":
                self.opt_sampler = [
                    s for s in samplers if type(s).__name__ == opt_sampler
                ][0]

        # Set fitted
        self.is_fitted = True

    def resample(self, x: pd.DataFrame, y: pd.Series) -> [pd.DataFrame, pd.Series]:
        """
        Resamples the data with the best sampler

        Parameters
        ----------
        x [pd.DataFrame]: Input data
        y [pd.Series]: Output data

        Returns
        -------
        x_res [pd.DataFrame]: Resampled input data
        y_res [pd.DataFrame]: Resampled output data
        """
        # Check whether fitted
        assert self.is_fitted, "Data Sampler not yet fitted. Run .fit(x, y) first. "

        # Check if we need sampling at all
        if self.opt_sampler is None:
            logger.info("No balancing needed :)")
            return x, y

        else:

            # Print start
            if self.verbosity > 0:
                logger.info(
                    "Resampling data with {}".format(type(self.opt_sampler).__name__)
                )

            # Resample
            x_res, y_res = self.opt_sampler.fit_resample(x, y)

            return x_res, y_res

    def fit_resample(self, x: pd.DataFrame, y: pd.Series):
        """
        Fits and returns resampled data

        Parameters
        ----------
        x [pd.DataFrame]: Input data
        y [pd.Series]: Output data

        Returns
        -------
        x_res [pd.DataFrame]: Resampled input data
        y_res [pd.Series]: Resampled output data
        """
        # Fit
        self.fit(x, y)

        # Resample
        x_res, y_res = self.resample(x, y)

        # Return
        return x_res, y_res
